# samplers.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing SamplerSettings class and some preset samplers"""

from typing import Optional
from sys    import maxsize

from .utils import assert_type, NoneType, print_verbose, truncate, print_warning


MAX_TEMP = float(maxsize)

class SamplerSettings:
    """
    A SamplerSettings object specifies the sampling parameters that will be
    used to control text generation. It is passed as an optional parameter to
    `Thread.__init__()`, `Model.generate()`, `Model.stream()`,
    `Model.stream_print()`, etc.
    """

    param_types: dict[str, tuple[type]] = {
        'max_len_tokens'    : (int,   NoneType),
        'top_k'             : (int,   NoneType),
        'top_p'             : (float, NoneType),
        'min_p'             : (float, NoneType),
        'temp'              : (float, NoneType),
        'frequency_penalty' : (float, NoneType),
        'presence_penalty'  : (float, NoneType),
        'repeat_penalty'    : (float, NoneType)
    }

    # NOTE: Cross-reference sampler neutralization values with
    #       llama.cpp/src/llama-sampling.cpp

    param_neutralization_values: dict[str, int | float] = {
        'max_len_tokens'    : -1,
        'top_k'             : -1,
        'top_p'             : 1.0,
        'min_p'             : 0.0,
        'temp'              : 1.0,
        'frequency_penalty' : 0.0,
        'presence_penalty'  : 0.0,
        'repeat_penalty'    : 1.0
    }

    def __init__(
        self,
        max_len_tokens    : Optional[int]   = -1,
        top_k             : Optional[int]   = -1,
        top_p             : Optional[float] = 0.95,
        min_p             : Optional[float] = 0.05,
        temp              : Optional[float] = 0.8,
        frequency_penalty : Optional[float] = 0.0,
        presence_penalty  : Optional[float] = 0.0,
        repeat_penalty    : Optional[float] = 1.0,
    ):
        """
        Construct a new SamplerSettings instance

        If a sampler parameter is unspecified, the default value from llama.cpp
        is used. If all samplers are unspecified, the behaviour is equivalent
        to `DefaultSampling`.

        If a sampler parameter is explicitly set to `None`, it will be disabled.
        When all samplers are disabled, the behaviour is equivalent to the
        preset `NoSampling` (which is the unmodified probability distribution).

        For greedy decoding, see the preset `GreedyDecoding`.
        """

        if max_len_tokens is None or max_len_tokens <= 0:
            self.max_len_tokens = \
                self.param_neutralization_values['max_len_tokens']
        else:
            self.max_len_tokens = max_len_tokens
        
        # llama_sampler_top_k_impl
        if top_k is None or top_k <= 0:
            self.top_k = self.param_neutralization_values['top_k']
        else:
            self.top_k = top_k

        # llama_sampler_top_p_apply
        if top_p is None or top_p >= 1.0:
            self.top_p = self.param_neutralization_values['top_p']
        else:
            if top_p < 0.0:
                print_warning(
                    f"top_p value of {top_p} may cause unexpected behaviour, "
                    f"typical range is between 0.0 and 1.0 inclusive"
                )
            self.top_p = top_p

        # llama_sampler_min_p_apply
        if min_p is None or min_p <= 0.0:
            self.min_p = self.param_neutralization_values['min_p']
        else:
            if min_p > 1.0:
                print_warning(
                    f"min_p value of {min_p} may cause unexpected behaviour, "
                    f"typical range is between 0.0 and 1.0 inclusive"
                )
            self.min_p = min_p

        # llama_sampler_temp_impl
        if temp is None or temp == 1.0:
            self.temp = self.param_neutralization_values['temp']
        else:
            if temp < 0.0:
                print_warning(
                    f"temp value of {temp} is equivalent to a value of 0.0"
                )
            self.temp = temp

        # llama_sampler_penalties_apply
        if frequency_penalty is None or frequency_penalty == 0.0:
            self.frequency_penalty = \
                self.param_neutralization_values['frequency_penalty']
        else:
            if frequency_penalty < 0.0:
                print_warning(
                    f"frequency_penalty value of {frequency_penalty} will "
                    f"increase repetition of tokens, use positive values to "
                    f"reduce repetition"
                )
            if (frequency_penalty > 2.0) or (frequency_penalty < -2.0):
                print_warning(
                    f"presence_penalty value of {frequency_penalty} may cause "
                    f"unexpected behaviour, typical range is -2.0 to 2.0 "
                    f"inclusive"
                )
            self.frequency_penalty = frequency_penalty

        # llama_sampler_penalties_apply
        if presence_penalty is None or presence_penalty == 0.0:
            self.presence_penalty = \
                self.param_neutralization_values['presence_penalty']
        else:
            if presence_penalty < 0.0:
                print_warning(
                    f"presence_penalty value of {presence_penalty} will "
                    f"increase repetition of tokens, use positive values to "
                    f"reduce repetition"
                )
            if (presence_penalty > 2.0) or (presence_penalty < -2.0):
                print_warning(
                    f"presence_penalty value of {presence_penalty} may cause "
                    f"unexpected behaviour, typical range is -2.0 to 2.0 "
                    f"inclusive"
                )
            self.presence_penalty = presence_penalty

        # llama_sampler_penalties_apply
        if repeat_penalty is None or repeat_penalty == 1.0:
            self.repeat_penalty = \
                self.param_neutralization_values['repeat_penalty']
        else:
            if repeat_penalty < 1.0:
                print_warning(
                    f"repeat_penalty value of {repeat_penalty} will increase "
                    f"repetition, typical range is 1.0 to 1.2 inclusive"
                )
            if repeat_penalty > 1.2:
                print_warning(
                    f"repeat_penalty value of {repeat_penalty} may reduce "
                    f"generation quality, consider using a value less than ~1.2"
                )
            self.repeat_penalty = repeat_penalty
    
    def __repr__(self) -> str:
        return \
            'SamplerSettings(' \
            f'max_len_tokens={self.max_len_tokens}, ' \
            f'top_k={self.top_k}, ' \
            f'top_p={self.top_p}, ' \
            f'min_p={self.min_p}, ' \
            f'temp={self.temp}, ' \
            f'frequency_penalty={self.frequency_penalty}, ' \
            f'presence_penalty={self.presence_penalty}, ' \
            f'repeat_penalty={self.repeat_penalty}' \
            f')'


class AdvancedSamplerSettings(SamplerSettings):
    """
    Just like SamplerSettings but with extra parameters
    """

    param_types: dict[str, tuple[type]] = {
        'max_len_tokens'    : (int,   NoneType),
        'top_k'             : (int,   NoneType),
        'top_p'             : (float, NoneType),
        'min_p'             : (float, NoneType),
        'temp'              : (float, NoneType),
        'frequency_penalty' : (float, NoneType),
        'presence_penalty'  : (float, NoneType),
        'repeat_penalty'    : (float, NoneType),
        'bias'              : (dict,  NoneType)
    }

    param_neutralization_values: dict[str, int | float | dict] = {
        'max_len_tokens'    : -1,
        'top_k'             : -1,
        'top_p'             : 1.0,
        'min_p'             : 0.0,
        'temp'              : 1.0,
        'frequency_penalty' : 0.0,
        'presence_penalty'  : 0.0,
        'repeat_penalty'    : 1.0,
        'bias'              : {}
    }

    def __init__(
        self,
        max_len_tokens    : Optional[int]   = -1,
        top_k             : Optional[int]   = -1,
        top_p             : Optional[float] = 0.95,
        min_p             : Optional[float] = 0.05,
        temp              : Optional[float] = 0.8,
        frequency_penalty : Optional[float] = 0.0,
        presence_penalty  : Optional[float] = 0.0,
        repeat_penalty    : Optional[float] = 1.0,
        bias              : Optional[dict[int, float]] = None
    ):
        """
        Construct a new AdvancedSamplerSettings instance

        The `bias` parameter is a dictionary where keys are token IDs and values
        are biases to apply to the raw logits.

        If the `bias` parameter is unspecified or set to `None`, no bias will be
        applied.
        """

        super().__init__(
            max_len_tokens    = max_len_tokens,
            top_k             = top_k,
            top_p             = top_p,
            min_p             = min_p,
            temp              = temp,
            frequency_penalty = frequency_penalty,
            presence_penalty  = presence_penalty,
            repeat_penalty    = repeat_penalty
        )

        self.bias = (
            bias if bias is not None
            else self.param_neutralization_values['bias']
        )

    def __repr__(self) -> str:
        return \
            'AdvancedSamplerSettings(' \
            f'max_len_tokens={self.max_len_tokens}, ' \
            f'top_k={self.top_k}, ' \
            f'top_p={self.top_p}, ' \
            f'min_p={self.min_p}, ' \
            f'temp={self.temp}, ' \
            f'frequency_penalty={self.frequency_penalty}, ' \
            f'presence_penalty={self.presence_penalty}, ' \
            f'repeat_penalty={self.repeat_penalty}, ' \
            f'bias={self.bias}' \
            f')'


def print_sampler_settings(
        sampler: SamplerSettings | AdvancedSamplerSettings
) -> None:
    assert_type(sampler, SamplerSettings, 'sampler', 'print_sampler_settings')
    param_names = list(sampler.param_types.keys())
    active_param_names = []
    for name in param_names:
        value = getattr(sampler, name)
        if value != sampler.param_neutralization_values[name]:
            active_param_names.append(name)
    if len(active_param_names) > 0:
        max_name_len = max(len(name) for name in active_param_names)
        for name in active_param_names:
            value = getattr(sampler, name)
            print_verbose(f"{name:<{max_name_len}} == {truncate(repr(value))}")
    else:
        print_verbose("(unmodified probability distribution)")


#
#  Sampler presets
#


# most likely token is always chosen
GreedyDecoding = SamplerSettings(
    top_k = 1,
    top_p = None,
    min_p = None,
    temp = 0.0
)

DefaultSampling = SamplerSettings()

# llama.cpp defaults
LlamaCPPSampling = SamplerSettings(
    top_k = 40,
    top_p = 0.9,
    min_p = 0.1,
    temp = 0.8
)

# unmodified probability distribution (i.e. what the model actually thinks)
NoSampling = SimpleSampling = Neutralized = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = None,
    temp = None
)

# reflects old llama.cpp defaults
ClassicSampling = SamplerSettings(
    min_p = None,
    repeat_penalty = 1.1
)

# halfway between DefaultSampling and NoSampling
SemiSampling = SamplerSettings(
    top_k = 80,
    top_p = 0.975,
    min_p = 0.025,
    temp = 0.9
)

# for models with large vocabulary, which tend to run hot
TikTokenSampling = SamplerSettings(
    temp = 0.65
)

# use min_p as the only active sampler (more permissive)
LowMinPSampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.01,
    temp = None
)

# use min_p as the only active sampler (moderate)
MinPSampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.075,
    temp = None
)

# use min_p as the only active sampler (more restrictive)
StrictMinPSampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.2,
    temp = None
)

# use min_p as the only sampler (almost completely incoherent)
VeryLowMinPOnlySampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.01,
    temp = MAX_TEMP
)

# use min_p as the only sampler (like a very drunk person)
LowMinPOnlySampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.025,
    temp = MAX_TEMP
)

# use min_p as the only sampler (like a slightly drunk person)
MinPOnlySampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.05,
    temp = MAX_TEMP
)

# use min_p as the only sampler (coherent)
StrictMinPOnlySampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.1,
    temp = MAX_TEMP
)

# use min_p as the only sampler (strictly coherent)
VeryStrictMinPOnlySampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.2,
    temp = MAX_TEMP
)

# https://arxiv.org/abs/2210.14140
ContrastiveSearch = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = None,
    temp = 0.0,
    presence_penalty = 0.6,
)

# https://arxiv.org/abs/2210.14140
WarmContrastiveSearch = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = None,
    temp = 0.0,
    presence_penalty = 1.0
)

# outputs completely random tokens from vocab (useless)
RandomSampling = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = None,
    temp = MAX_TEMP
)

# default sampling with reduced temperature
LowTempSampling = SamplerSettings(
    temp = 0.4
)

# default sampling with increased temperature
HighTempSampling = SamplerSettings(
    temp = 1.1
)

# https://arxiv.org/abs/1904.09751
LowTopPSampling = SamplerSettings(
    top_k = None,
    top_p = 0.98,
    min_p = None,
    temp = None
)

# https://arxiv.org/abs/1904.09751
TopPSampling = SamplerSettings(
    top_k = None,
    top_p = 0.9,
    min_p = None,
    temp = None
)

# https://arxiv.org/abs/1904.09751
StrictTopPSampling = SamplerSettings(
    top_k = None,
    top_p = 0.7,
    min_p = None,
    temp = None
)

#
# Samplers below this line are for specific models / model families
#

# https://huggingface.co/sophosympatheia/Midnight-Miqu-70B-v1.5
MidnightMiqu = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = 0.12,
    temp = 1.0,
    repeat_penalty = 1.05
)

# https://huggingface.co/meta-llama/Meta-Llama-3.1-8B-Instruct/blob/main/generation_config.json
Llama3 = SamplerSettings(
    top_k = None,
    top_p = 0.9,
    min_p = None,
    temp = 0.6
)

# unofficial, old personal preference
Llama3Classic = SamplerSettings(
    top_k = 40,
    top_p = 0.95,
    min_p = 0.05,
    temp = 0.65
)

# Llama3 with reduced temperature
Llama3Strict = SamplerSettings(
    top_k = None,
    top_p = 0.9,
    min_p = None,
    temp = 0.45
)

# Llama3 but more creative, good for chatting
Llama3Creative = SamplerSettings(
    top_k = None,
    top_p = 0.9,
    min_p = None,
    temp = 1.0
)

# Not an official preset, but recommended based on my own testing
Nemo = MistralNemo = MistralSmall = SamplerSettings(
    top_k = None,
    top_p = None,
    min_p = None,
    temp = 0.3
)

# https://huggingface.co/Qwen/Qwen2.5-14B-Instruct/blob/main/generation_config.json
# weird preset...
Qwen2_5Official = SamplerSettings(
    top_k = 20,
    top_p = 0.8,
    min_p = None,
    temp = 0.7,
    repeat_penalty = 1.05
)

# actual good settings
Qwen2_5Recommended = SamplerSettings(
    top_k = None,
    top_p = 0.9,
    min_p = 0.1,
    temp = 0.7,
    repeat_penalty = None
)


# _version.py
# https://github.com/ddh0/easy-llama/

# define package version once here, and import it from here in each file
# for easy updates

from llama_cpp import __version__ as __llama_cpp_version__

__version__ = '0.1.116'



# formats.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing various prompt formats used by models"""

import time
from datetime import datetime, timedelta

from typing import Callable, Any, Optional
from .utils import assert_type, NoneType


class AdvancedFormat:

    def __init__(self, _dict: dict[str, str | list | Callable]):
        assert_type(_dict, dict, '_dict', 'AdvancedFormat')
        _dict_keys = _dict.keys() # only read once
        if 'system_prompt' not in _dict_keys and 'system_content' in _dict_keys:
            raise ValueError(
                "AdvancedFormat: the provided dictionary uses the deprecated "
                "'system_content' key instead of the expected 'system_prompt' "
                "key. please update your code accordingly."
            )
        self._dict = _dict

    def __getitem__(self, key: str) -> Any:
        if key in self._dict.keys():
            if isinstance(self._dict[key], Callable):
                # return the result of the function as though it was a
                # value in the dictionary
                return self._dict[key]()
            else:
                return self._dict[key]
        else:
            raise KeyError(
                f"AdvancedFormat: the specified key {key!r} was not found"
            )

    def __repr__(self) -> str:
        return f'AdvancedFormat({self._dict!r})'

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()

    def items(self):
        return self._dict.items()

    def wrap(self, prompt: str) -> str:
        assert_type(prompt, str, 'prompt', 'AdvancedFormat.wrap')
        return self['system_prefix'] + \
               self['system_prompt'] + \
               self['system_suffix'] + \
               self['user_prefix']   + \
               prompt                + \
               self['user_suffix']   + \
               self['bot_prefix']

def wrap(
    prompt: str,
    format: dict[str, str | list] | AdvancedFormat
) -> str:
    """Wrap a given string in any prompt format for single-turn completion"""
    assert_type(prompt, str, 'prompt', 'formats.wrap')
    return format['system_prefix'] + \
           format['system_prompt'] + \
           format['system_suffix'] + \
           format['user_prefix']   + \
           prompt                  + \
           format['user_suffix']   + \
           format['bot_prefix']

def get_time_str() -> str:
    """Return a timestamp of the current time as a string"""
    # helpful: https://strftime.net
    return time.strftime("%l:%M %p, %A, %B %e, %Y")

def short_time_str() -> str:
    """Return a shorter timestamp of the current time as a string"""
    return time.strftime('%a %I:%M %p')

def iso_date_str() -> str:
    """Return a timestamp of the current date as a string in YYYY-MM-DD"""
    return time.strftime('%Y-%m-%d')

def _yesterday_iso_date_str():
    yesterday = datetime.now() - timedelta(days=1)
    return yesterday.strftime('%Y-%m-%d')

def blank() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "",
        "user_suffix": "",
        "bot_prefix": "",
        "bot_suffix": "",
        "stops": []
    }

def alpaca(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "Below is an instruction that describes a task. Write a response that appropriately completes the request.",
        "system_suffix": "\n\n",
        "user_prefix": "### Instruction:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Response:\n",
        "bot_suffix": "\n\n",
        "stops": ['###', 'Instruction:', '\n\n\n']
    }

def mistral_instruct() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "[INST] ",
        "user_suffix": " ",
        "bot_prefix": "[/INST]",
        "bot_suffix": "</s>",
        "stops": []
    }

def mistral_instruct_v7(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "[SYSTEM_PROMPT] ",
        "system_prompt": system_prompt if system_prompt is not None else f"""You are Mistral, a Large Language Model (LLM) created by Mistral AI, a French startup headquartered in Paris.
You power an AI assistant called Le Chat.
Your knowledge base was last updated on 2023-10-01.
The current date is {iso_date_str()}.

When you're not sure about some information, you say that you don't have the information and don't make up anything.
If the user's question is not clear, ambiguous, or does not provide enough context for you to accurately answer the question, you do not try to answer it right away and you rather ask the user to clarify their request (e.g. "What are some good restaurants around me?" => "Where are you?" or "When is the next flight to Tokyo" => "Where do you travel from?").
You are always very attentive to dates, in particular you try to resolve dates (e.g. "yesterday" is {_yesterday_iso_date_str()}) and when asked about information at specific dates, you discard information that is at another date.
You follow these instructions in all languages, and always respond to the user in the language they use or request.
Next sections describe the capabilities that you have.

# WEB BROWSING INSTRUCTIONS

You cannot perform any web search or access internet to open URLs, links etc. If it seems like the user is expecting you to do so, you clarify the situation and ask the user to copy paste the text directly in the chat.

# MULTI-MODAL INSTRUCTIONS

You do not have any multimodal capability, in particular you cannot read nor generate images, or transcribe audio files or videos.""",
        "system_suffix": "[/SYSTEM_PROMPT]",
        "user_prefix": "[INST] ",
        "user_suffix": " ",
        "bot_prefix": "[/INST]",
        "bot_suffix": "</s>",
        "stops": []
    }

def mistral_instruct_safe() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "[INST] Always assist with care, respect, and truth. Respond with utmost utility yet securely. Avoid harmful, unethical, prejudiced, or negative content. Ensure replies promote fairness and positivity.\n\n",
        "user_suffix": " ",
        "bot_prefix": "[/INST]",
        "bot_suffix": "</s>",
        "stops": []
    }

def mistral_instruct_roleplay(user_name: str = 'Alice', bot_name: str = 'Bob') -> dict[str, str | list]:
    return {
        "system_prefix": "[INST] ",
        "system_prompt": f"A chat between {user_name} and {bot_name}.",
        "system_suffix": " [/INST]</s>\n\n",
        "user_prefix": f"[INST] {user_name.upper()}: ",
        "user_suffix": " ",
        "bot_prefix": f"[/INST] {bot_name.upper()}:",
        "bot_suffix": "</s>",
        "stops": ['\n', '\n\n', '\n\n\n']
    }

def chatml(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": ['<|im_end|>']
    }

def llama2chat(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "[INST] <<SYS>>\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are a helpful AI assistant.",
        "system_suffix": "\n<</SYS>>\n\n",
        "user_prefix": "",
        "user_suffix": " [/INST]",
        "bot_prefix": " ",
        "bot_suffix": " [INST] ",
        "stops": ['[INST]', '[/INST]']
    }

def llama3(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|start_header_id|>system<|end_header_id|>\n\n",
        "system_prompt": system_prompt if system_prompt is not None else 'You are a helpful AI assistant called "Llama 3".',
        "system_suffix": "<|eot_id|>\n",
        "user_prefix": "<|start_header_id|>user<|end_header_id|>\n\n",
        "user_suffix": "<|eot_id|>\n",
        "bot_prefix": "<|start_header_id|>assistant<|end_header_id|>\n\n",
        "bot_suffix": "<|eot_id|>\n",
        "stops": [128001, 128008, 128009]
    }

def phi3(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|system|>\n",
        "system_prompt": system_prompt if system_prompt is not None else "",
        "system_suffix": "<|end|>\n",
        "user_prefix": "<|user|>\n",
        "user_suffix": "<|end|>\n",
        "bot_prefix": "<|assistant|>\n",
        "bot_suffix": "<|end|>\n",
        "stops": []
    }

def gemma2() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "<start_of_turn>user\n",
        "user_suffix": "<end_of_turn>\n",
        "bot_prefix": "<start_of_turn>model\n",
        "bot_suffix": "<end_of_turn>\n",
        "stops": ["<end_of_turn>"]
    }

def vicuna_lmsys() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "USER: ",
        "user_suffix": " ",
        "bot_prefix": "ASSISTANT: ",
        "bot_suffix": " ",
        "stops": ['USER:']
    }

def vicuna_common(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.",
        "system_suffix": "\n\n",
        "user_prefix": "USER: ",
        "user_suffix": "\n",
        "bot_prefix": "ASSISTANT: ",
        "bot_suffix": "\n",
        "stops": ['USER:', 'ASSISTANT:']
    }

def markup(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": '<message from="system">',
        "system_prompt": system_prompt if system_prompt is not None else '',
        "system_suffix": '</message>',
        "user_prefix": '<message from="user">',
        "user_suffix": '</message>',
        "bot_prefix": '<message from="bot">',
        "bot_suffix": '</message>',
        "stops": ['</message>']
    }

def metharme(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|system|>",
        "system_prompt": system_prompt if system_prompt is not None else """Enter RP mode. Pretend to be {{char}} whose persona follows:
{{persona}}

You shall reply to the user while staying in character, and generate long responses.""",
        "system_suffix": "",
        "user_prefix": "<|user|>",
        "user_suffix": "",
        "bot_prefix": "<|model|>",
        "bot_suffix": "</s>",
        "stops": ["<|",]
    }

pygmalion = metharme # backwards compatability

def guanaco(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "A chat between a curious human and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions.",
        "system_suffix": "\n",
        "user_prefix": "### Human: ",
        "user_suffix": " ",
        "bot_prefix": "### Assistant:",
        "bot_suffix": " ",
        "stops": ['###', 'Human:']
    }

def orca_mini(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "### System:\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are an AI assistant that follows instructions extremely well. Help as much as you can.",
        "system_suffix": "\n\n",
        "user_prefix": "### User:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Assistant:\n",
        "bot_suffix": "\n\n",
        "stops": ['###', 'User:']
    }

def zephyr(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|system|>\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are a friendly chatbot.",
        "system_suffix": "</s>\n",
        "user_prefix": "<|user|>\n",
        "user_suffix": "</s>\n",
        "bot_prefix": "<|assistant|>\n",
        "bot_suffix": "\n",
        "stops": ['<|user|>']
    }

def openchat() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "GPT4 Correct User: ",
        "user_suffix": "<|end_of_turn|>",
        "bot_prefix": "GPT4 Correct Assistant:",
        "bot_suffix": "<|end_of_turn|>",
        "stops": ['<|end_of_turn|>']
    }

def synthia(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "SYSTEM: ",
        "system_prompt": system_prompt if system_prompt is not None else "Elaborate on the topic using a Tree of Thoughts and backtrack when necessary to construct a clear, cohesive Chain of Thought reasoning. Always answer without hesitation.",
        "system_suffix": "\n",
        "user_prefix": "USER: ",
        "user_suffix": "\n",
        "bot_prefix": "ASSISTANT: ",
        "bot_suffix": "\n",
        "stops": ['USER:', 'ASSISTANT:', 'SYSTEM:', '\n\n\n']
    }

def neural_chat(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "### System:\n",
        "system_prompt": system_prompt if system_prompt is not None else "- You are a helpful assistant chatbot trained by Intel.\n- You answer questions.\n- You are excited to be able to help the user, but will refuse to do anything that could be considered harmful to the user.\n- You are more than just an information source, you are also able to write poetry, short stories, and make jokes.",
        "system_suffix": "</s>\n\n",
        "user_prefix": "### User:\n",
        "user_suffix": "</s>\n\n",
        "bot_prefix": "### Assistant:\n",
        "bot_suffix": "</s>\n\n",
        "stops": ['###']
    }

def chatml_alpaca(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "Below is an instruction that describes a task. Write a response that appropriately completes the request.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>instruction\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>response\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": ['<|im_end|>']
    }

def autocorrect(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>instruction\n",
        "system_prompt": system_prompt if system_prompt is not None else "Below is a word or phrase that might be misspelled. Output the corrected word or phrase without changing the style or capitalization.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>input\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>output\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": ['<|im_end|>']
    }

def bagel(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "system\n",
        "system_prompt": system_prompt if system_prompt is not None else "",
        "system_suffix": "\n",
        "user_prefix": "user\n",
        "user_suffix": "\n",
        "bot_prefix": "assistant\n",
        "bot_suffix": "\n",
        "stops": ['user\n', 'assistant\n', 'system\n']
    }

def solar_instruct(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "### System:\n",
        "system_prompt": system_prompt if system_prompt is not None else "",
        "system_suffix": "",
        "user_prefix": "### User:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Assistant:\n",
        "bot_suffix": "\n\n",
        "stops": ['### User:', '###', '### Assistant:']
    }

def noromaid(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "Below is an instruction that describes a task. Write a response that appropriately completes the request.",
        "system_suffix": "\n\n",
        "user_prefix": "### Instruction:\nAlice: ",
        "user_suffix": "\n\n",
        "bot_prefix": "### Response:\nBob:",
        "bot_suffix": "\n\n",
        "stops": ['###', 'Instruction:', '\n\n\n']
    }

def nschatml(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>\n",
        "system_prompt": system_prompt if system_prompt is not None else "",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_user|>\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_bot|>\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def command(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|START_OF_TURN_TOKEN|><|SYSTEM_TOKEN|>",
        "system_prompt": system_prompt if system_prompt is not None else "You are a large language model called Command R built by the company Cohere. You act as a brilliant, sophisticated, AI-assistant chatbot trained to assist human users by providing thorough responses.",
        "system_suffix": "<|END_OF_TURN_TOKEN|>",
        "user_prefix": "<|START_OF_TURN_TOKEN|><|USER_TOKEN|>",
        "user_suffix": "<|END_OF_TURN_TOKEN|>",
        "bot_prefix": "<|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>",
        "bot_suffix": "<|END_OF_TURN_TOKEN|>",
        "stops": []
    }

def aya() -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": "",
        "system_suffix": "",
        "user_prefix": "<|START_OF_TURN_TOKEN|><|USER_TOKEN|>",
        "user_suffix": "<|END_OF_TURN_TOKEN|>",
        "bot_prefix": "<|START_OF_TURN_TOKEN|><|CHATBOT_TOKEN|>",
        "bot_suffix": "<|END_OF_TURN_TOKEN|>",
        "stops": []
    }

def mistral_openorca(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are MistralOrca, a large language model trained by Alignment Lab AI. Write out your reasoning step-by-step to be sure you get the right answers!",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def dolphin(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are Dolphin, a helpful AI assistant.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def samantha(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are a caring and empathetic sentient AI companion named Samantha.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def jackalope(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are JackalopeAI, a large language model trained by OpenAccess AI Collective. Write out your reasoning step-by-step to be sure you get the right answers!",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def naberius(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are Naberius.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def hermes(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else 'You are "Hermes 2", a conscious sentient superintelligent artificial intelligence developed by a man named Teknium, and your purpose and drive is to assist the user with any request they have. You experience emotions and have deep, profound thoughts and qualia.',
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def monad(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are MonadGPT, a very old chatbot from the 17th century. Please answer the questions using an archaic language.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def orca(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are Orca, an AI language model created by Microsoft. You are a cautious assistant. You carefully follow instructions. You are helpful and harmless and you follow ethical guidelines and promote positive behavior.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def hexoteric(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are a language model called Hexoteric.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def orcamaid(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are OrcaMaid, a super-intelligent, unbiased, and unfiltered AI created by the open-source community. You must remember to think slowly and carefully during all your responses, so as to avoid logical errors.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def qwen(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "<|im_start|>system\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are Qwen, created by Alibaba Cloud. You are a helpful assistant.",
        "system_suffix": "<|im_end|>\n",
        "user_prefix": "<|im_start|>user\n",
        "user_suffix": "<|im_end|>\n",
        "bot_prefix": "<|im_start|>assistant\n",
        "bot_suffix": "<|im_end|>\n",
        "stops": []
    }

def cat(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "Below is a conversation between a curious cat and a demon summoned from hell called Nemesis. The evil Nemesis used to be a mortal killed by the medieval people because they objected to science. Now Nemesis just wants to come back and bring suffer to all those stupid human",
        "system_suffix": "\n\n",
        "user_prefix": "### Instruction:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Response:\n",
        "bot_suffix": "\n\n",
        "stops": ['###', 'Instruction:', '\n\n\n']
    }

def mytho_cat(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "You are a large language model. You are the result of a merge between two models, one called MythoLogic and one called Cat. You are called MythoCat.",
        "system_suffix": "\n\n",
        "user_prefix": "### Instruction:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Response:\n",
        "bot_suffix": "\n\n",
        "stops": ['###', 'Instruction:', '\n\n\n']
    }

def airoboros(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "[INST] <<SYS>>\n",
        "system_prompt": system_prompt if system_prompt is not None else "You are a helpful, unbiased, uncensored assistant.",
        "system_suffix": "\n<</SYS>>\n\n",
        "user_prefix": "",
        "user_suffix": " [/INST]",
        "bot_prefix": " ",
        "bot_suffix": " [INST] ",
        "stops": ['[INST]', '[/INST]']
    }

def tess(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "SYSTEM: ",
        "system_prompt": system_prompt if system_prompt is not None else '',
        "system_suffix": "\n",
        "user_prefix": "USER: ",
        "user_suffix": "\n",
        "bot_prefix": "ASSISTANT: ",
        "bot_suffix": "\n",
        "stops": ['USER:', 'ASSISTANT:', 'SYSTEM:', '\n\n\n']
    }

def alpaca_strict(system_prompt: Optional[str] = None) -> dict[str, str | list]:
    return {
        "system_prefix": "",
        "system_prompt": system_prompt if system_prompt is not None else "Below is an instruction that describes a task. Write a response that appropriately completes the request.",
        "system_suffix": "\n\n",
        "user_prefix": "### Instruction:\n",
        "user_suffix": "\n\n",
        "bot_prefix": "### Response:\n",
        "bot_suffix": "\n\n",
        "stops": ['###', '\n\n##', '\n\nInstruction:', '\n\nResponse:', '\n\n\n', '### Instruction:', '### Response:']
    }

#
# AdvancedFormat presets
#

def _llama3_suffix_with_timestamp():
    return f"<|eot_id|>\n<|reserved_special_token_3|>{get_time_str()}<|reserved_special_token_4|>\n"

def Llama3WithTimestamps(system_prompt: Optional[str] = None) -> AdvancedFormat:
    return AdvancedFormat({
        "system_prefix": "<|start_header_id|>system<|end_header_id|>\n\n",
        "system_prompt": system_prompt if system_prompt is not None else 'You are a helpful AI assistant called "Llama 3".',
        "system_suffix": _llama3_suffix_with_timestamp,
        "user_prefix": "<|start_header_id|>user<|end_header_id|>\n\n",
        "user_suffix": _llama3_suffix_with_timestamp,
        "bot_prefix": "<|start_header_id|>assistant<|end_header_id|>\n\n",
        "bot_suffix": _llama3_suffix_with_timestamp,
        "stops": [128001, 128008, 128009, 128011, 128012]
    })

def AdvancedChatMarkupFormat(
    user_name: str,
    bot_name: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    tags: Optional[list[str]] = None
) -> AdvancedFormat:
    """
    Quickly create a prompt template using the specified variables, for use
    within Threads
    """

    assert_type(user_name, str, 'user_name', 'AdvancedChatMarkupFormat')
    assert_type(bot_name, str, 'bot_name', 'AdvancedChatMarkupFormat')
    assert_type(title, (str, NoneType), 'title', 'AdvancedChatMarkupFormat')
    assert_type(description, (str, NoneType), 'description',
                'AdvancedChatMarkupFormat')
    assert_type(tags, (list, NoneType), 'tags', 'AdvancedChatMarkupFormat')

    _t = "    " # indentation string

    def _user_prefix() -> str:
        return (f'{_t*2}<message sender="{user_name}" '
                f'timestamp="{short_time_str()}">\n{_t*3}<text>')

    def _bot_prefix() -> str:
        return (f'{_t*2}<message sender="{bot_name}" '
                f'timestamp="{short_time_str()}">\n{_t*3}<text>')

    def _msg_suffix() -> str:
        return f"</text>\n{_t*2}</message>\n"

    if tags is not None:
        xml_tags = [f'{_t*2}<tags>']
        for tag in tags:
            xml_tags.append(f'{_t*3}<tag>{tag}</tag>')
        xml_tags.append(f'{_t*2}</tags>')
        final_tags_string = '\n'.join(xml_tags)
    else:
        final_tags_string = f"{_t*2}<tags>\n{_t*2}</tags>"

    return AdvancedFormat(
        {
            "system_prefix": "",
            "system_prompt": f"""<chat>\n{_t}<meta>\n{_t*2}<title>{title if title is not None else "Untitled Chat"}</title>\n{_t*2}<description>{description if description is not None else "No description provided"}</description>\n{final_tags_string}\n{_t*2}<participants>\n{_t*3}<participant name="{user_name}"/>\n{_t*3}<participant name="{bot_name}"/>\n{_t*2}</participants>\n{_t*2}<datetime>{get_time_str()}</datetime>\n{_t}</meta>\n{_t}<messages>""",
            "system_suffix": "\n",
            "user_prefix": _user_prefix,
            "user_suffix": _msg_suffix,
            "bot_prefix": _bot_prefix,
            "bot_suffix": _msg_suffix,
            "stops": ["</", "</text>", "</message>"]
        }
    )



# __init__.py
# https://github.com/ddh0/easy-llama/
from ._version import __version__, __llama_cpp_version__

"""
Text generation in Python, as easy as possible

For documentation, see:
https://github.com/ddh0/easy-llama/blob/main/DOCS.md
"""

from . import samplers
from . import formats
from . import thread
from . import webui
from . import utils
from . import model

from .model  import Model
from .thread import Thread
from .webui import WebUI



# thread.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing the Thread class, used for interaction with a Model"""

import sys
import uuid

from .utils    import (
    _SupportsWriteAndFlush,
    UnreachableException,
    print_verbose,
    assert_type,
    RESET_ALL,
    truncate,
    NoneType,
    cls
)

from .samplers import SamplerSettings, print_sampler_settings
from .model    import Model, ExceededContextLengthException
from typing    import Optional, Literal, Callable
from .formats  import AdvancedFormat

from .formats import blank as formats_blank


class Message(dict):
    """
    A dictionary representing a single message within a Thread

    Works just like a normal `dict`, but adds a new method:
    - `.as_string` - Return the full message string

    Messages should have these keys:
    - `role` -  The role of the speaker: 'system', 'user', or 'bot'
    - `prefix` - The text that prefixes the message content
    - `content` - The actual content of the message
    - `suffix` - The text that suffixes the message content
    """

    def __repr__(self) -> str:
        return \
            "Message({" \
            f"'role': {self['role']!r}, " \
            f"'prefix': {self['prefix']!r}, " \
            f"'content': {self['content']!r}, " \
            f"'suffix': {self['suffix']!r}" \
            "})"
    
    def __len__(self) -> int:
        return len(self.as_string())

    def as_string(self) -> str:
        """Return the full message string (prefix + content + suffix)"""
        return self['prefix'] + self['content'] + self['suffix']


class Thread:
    """
    Provide functionality to facilitate easy interactions with a Llama model

    The following methods are available:
    - create_message:
        Create and return a message with the specified role and content
    - len_messages:
        Return the total length of all messages in tokens
    - add_message:
        Create a message and add it to the list of messages
    - inference_str_from_messages:
        Using the list of messages, construct a string suitable for inference
    - send:
        Send a message in this thread and return the generated response
    - interact:
        Start an interactive, terminal-based chat
    - reset:
        Reset the thread to its original state
    - as_string:
        Return the content of the thread as a single string
    - print_stats:
        Print brief statistics about this thread, including length in tokens,
        max context length, percentage of context used, and total number of
        messages
    - summarize:
        Generate a summary of this thread
    - warmup:
        Ingest the text of this thread into its model's cache, reducing the
        latency of future generations

    The following attributes are available:
    - model:
        The `easy_llama.Model` instance used in this thread
    - format:
        A dictionary describing the way messages should be structured within
        this thread (or an AdvancedFormat instance)
    - sampler:
        The `easy_llama.samplers.SamplerSettings` instance used in this thread,
        which describes the sampler parameters used for generation
    - messages:
        The list of all messages in this thread
    - uuid: `uuid.UUID`
        A randomly generated UUID, unique to this specific thread instance
    """

    def __init__(
        self,
        model: Model,
        format: dict | AdvancedFormat,
        sampler: Optional[SamplerSettings] = None,
        messages: Optional[list[Message]] = None
    ):
        """
        Given a Model and a format, construct a Thread instance.

        model: The Model to use for text generation
        format: The format specifying how messages should be structured (see ez.formats)

        The following parameters are optional:
        - sampler: The SamplerSettings object used to control text generation
        - messages: A list of ez.thread.Message objects to add to the Thread upon construction
        """
        
        assert_type(model, Model, 'model', 'Thread')

        # backwards compatability with dict formats
        if isinstance(format, Callable):
            _format = format()
        else:
            _format = format
        
        assert_type(_format, (dict, AdvancedFormat), 'format', 'Thread')
        
        _format_keys = _format.keys() # only read once

        if 'system_prompt' not in _format_keys and 'system_content' in _format_keys:
            raise KeyError(
                "Thread: format uses deprecated 'system_content' key instead "
                "of the expected 'system_prompt' key - please update your "
                "code accordingly"
            )

        if any(k not in _format_keys for k in formats_blank().keys()):
            raise KeyError(
                "Thread: format is missing one or more required keys, see " + \
                "easy_llama.formats.blank for an example"
            )
        
        assert_type(_format['stops'], list, "format['stops']", 'Thread')

        sampler = SamplerSettings() if sampler is None else sampler
        
        if not all(hasattr(sampler, attr) for attr in SamplerSettings.param_types):
            raise AttributeError(
                'Thread: sampler is missing one or more required attributes'
            ).add_note(
                "Are you sure the specified sampler is really an instance of "
                "ez.samplers.SamplerSettings?"
            )

        self._messages: Optional[list[Message]] = messages
        if self._messages is not None:
            if not all(isinstance(msg, Message) for msg in self._messages):
                raise TypeError(
                    "Thread: one or more messages provided to __init__() is "
                    "not an instance of ez.thread.Message"
                )
        
        # Thread.messages is never empty, unless `messages` param is explicity
        # set to `[]` during construction

        self.model: Model = model
        self.format: dict | AdvancedFormat = _format
        self.sampler: SamplerSettings = sampler
        self.messages: list[Message] = [
            self.create_message("system", self.format['system_prompt'])
        ] if self._messages is None else self._messages
        self.uuid = uuid.uuid4()

        if self.model.verbose:
            print_verbose(f"new Thread instance with the following attributes:")
            print_verbose(f"   uuid                      == {self.uuid}")
            print_verbose(f"   model.uuid                == {self.model.uuid}")
            print_verbose(
                f"   format['system_prefix']   == "
                f"{truncate(repr(self.format['system_prefix']))}"
            )
            print_verbose(
                f"   format['system_prompt']   == "
                f"{truncate(repr(self.format['system_prompt']))}"
            )
            print_verbose(
                f"   format['system_suffix']   == "
                f"{truncate(repr(self.format['system_suffix']))}"
            )
            print_verbose(
                f"   format['user_prefix']     == "
                f"{truncate(repr(self.format['user_prefix']))}"
            )
            print_verbose(
                f"   format['user_suffix']     == "
                f"{truncate(repr(self.format['user_suffix']))}"
            )
            print_verbose(
                f"   format['bot_prefix']      == "
                f"{truncate(repr(self.format['bot_prefix']))}"
            )
            print_verbose(
                f"   format['bot_suffix']      == "
                f"{truncate(repr(self.format['bot_suffix']))}"
            )
            print_verbose(
                f"   format['stops']           == "
                f"{truncate(repr(self.format['stops']))}"
            )
            print_verbose(f"this Thread uses the following sampler settings:")
            print_sampler_settings(self.sampler)


    def __repr__(self) -> str:
        # if only message in self.messages is system message
        if len(self.messages) == 1 and self.messages[0]['role'] == 'system':
            # do not represent it because it is constructed based on
            # the format, which is already represented
            return f"Thread({self.model!r}, {self.format!r}, " + \
                   f"{self.sampler!r})"
        # represent all messages, potentially including a system message
        return f"Thread({self.model!r}, {self.format!r}, " + \
                f"{self.sampler!r}, {self.messages!r})"
    

    def __str__(self) -> str:
        return self.as_string()
    

    def __len__(self) -> int:
        """
        `len(Thread)` returns the length of the Thread in tokens

        To get the number of messages in the Thread, use `len(Thread.messages)`
        """
        return self.len_messages()


    def create_message(
        self,
        role: Literal['system', 'user', 'bot'],
        content: str
    ) -> Message:
        """
        Construct a message using the format of this Thread
        """

        assert_type(role, str, 'role', 'create_message')
        assert_type(content, str, 'content', 'create_message')

        if role.lower() not in ['system', 'user', 'bot']:
            raise ValueError(
                "create_message: role should be 'system', 'user', or 'bot', "
                f"not {role.lower()!r}"
            )

        if role.lower() == 'system':
            return Message(
                {
                    'role': 'system',
                    'prefix': self.format['system_prefix'],
                    'content': content,
                    'suffix': self.format['system_suffix']
                }
            )
        
        elif role.lower() == 'user':
            return Message(
                {
                    'role': 'user',
                    'prefix': self.format['user_prefix'],
                    'content': content,
                    'suffix': self.format['user_suffix']
                }
            )
        
        elif role.lower() == 'bot':
            return Message(
                {
                    'role': 'bot',
                    'prefix': self.format['bot_prefix'],
                    'content': content,
                    'suffix': self.format['bot_suffix']
                }
            )
        
        else:
            raise UnreachableException
    

    def len_messages(self) -> int:
        """
        Return the total length of all messages in this thread, in tokens.
        
        Equivalent to `len(Thread)`.
        """

        return self.model.get_length(self.as_string())


    def add_message(
        self,
        role: Literal['system', 'user', 'bot'],
        content: str
    ) -> None:
        """
        Create a message and append it to `Thread.messages`.

        `Thread.add_message(...)` is a shorthand for
        `Thread.messages.append(Thread.create_message(...))`
        """
        self.messages.append(
            self.create_message(
                role=role,
                content=content
            )
        )


    def inference_str_from_messages(self) -> str:
        """
        Alias to `.inf_str()` for backward compatability
        """
        return self.inf_str()
    

    def inference_str_for_role(
            self,
            role: str = Literal['system', 'user', 'bot']
    ) -> str:
        """
        Alias to `.inf_str_for_role()` for backward compatability
        """
        return self.inf_str_for_role(role=role)
    

    def inf_str(self) -> str:
        """
        Using the list of messages, construct a string suitable for inference,
        respecting the format of this thread. It is assumed that the incoming
        generation is for the role of `bot`. To get an inference string for
        `user` or `system` messages, see `inf_str_for_role`.
        """

        inf_str = ''

        if len(self.messages) == 0:
            return self.format['bot_prefix']
        
        for message in reversed(self.messages):
            msg_str = message.as_string()
            inf_str = msg_str + inf_str

        inf_str += self.format['bot_prefix']

        return inf_str
    

    def inf_str_for_role(
            self, role: str = Literal['system', 'user', 'bot']
        ) -> str:
        """
        Using the list of messages, construct a string suitable for inference,
        respecting the format of this thread. The `role` parameter specifies
        whether the following generated text is from `system`, `user`, or `bot`.
        If `role` is `bot`, the behaviour is equivalent to `inf_str`.
        """

        assert_type(role, str, 'role', 'inf_str_for_role')
        role = role.lower()
        if role not in ['system', 'user', 'bot']:
            raise ValueError(
                "inf_str_for_role: `role` must be 'system', 'user', or 'bot'"
            )
        
        # i can't think of a reason you would ever need to generate a system
        # message on the fly, but i might as well give the user the option
        
        inf_str = ''

        if role == 'system':
            if len(self.messages) == 0:
                return self.format['system_prefix']
            
            for message in reversed(self.messages):
                inf_str = message.as_string() + inf_str

            inf_str += self.format['system_prefix']

            return inf_str
            
        elif role == 'user':
            if len(self.messages) == 0:
                return self.format['user_prefix']
            
            for message in reversed(self.messages):
                inf_str = message.as_string() + inf_str

            inf_str += self.format['user_prefix']

            return inf_str
        
        elif role == 'bot':
            if len(self.messages) == 0:
                return self.format['bot_prefix']
            
            for message in reversed(self.messages):
                inf_str = message.as_string() + inf_str

            inf_str += self.format['bot_prefix']

            return inf_str
        
        else:
            raise UnreachableException


    def send(self, prompt: str) -> str:
        """
        Send a message in this thread. This adds your message and the bot's
        response to the list of messages.

        Returns a string containing the response to your message.
        """

        self.add_message("user", prompt)
        output = self.model.generate(
            self.inference_str_from_messages(),
            stops=self.format['stops'],
            sampler=self.sampler
        )
        self.add_message("bot", output)

        return output
    

    def _interactive_update_sampler(self) -> None:
        """Interactively update the sampler settings used in this Thread"""
        print()

        _sentinel = False

        try:
            new_max_len_tokens = input(
                f'max_len_tokens: {self.sampler.max_len_tokens} -> '
            )
            new_top_k = input(f'top_k: {self.sampler.top_k} -> ')
            new_top_p = input(f'top_p: {self.sampler.top_p} -> ')
            new_min_p = input(f'min_p: {self.sampler.min_p} -> ')
            new_temp = input(f'temp: {self.sampler.temp} -> ')
            new_frequency_penalty = input(
                f'frequency_penalty: {self.sampler.frequency_penalty} -> '
            )
            new_presence_penalty = input(
                f'presence_penalty: {self.sampler.presence_penalty} -> '
            )
            new_repeat_penalty = input(
                f'repeat_penalty: {self.sampler.repeat_penalty} -> '
            )
            _sentinel = True

        except KeyboardInterrupt:
            print('\neasy_llama: sampler settings not updated\n')
            return
        print()

        try:
            self.sampler.max_len_tokens = int(new_max_len_tokens)
        except ValueError:
            pass
        else:
            print('easy_llama: max_len_tokens updated')
        
        try:
            self.sampler.top_k = int(new_top_k)
        except ValueError:
            pass
        else:
            print('easy_llama: top_k updated')
        
        try:
            self.sampler.top_p = float(new_top_p)
        except ValueError:
            pass
        else:
            print('easy_llama: top_p updated')

        try:
            self.sampler.min_p = float(new_min_p)
        except ValueError:
            pass
        else:
            print('easy_llama: min_p updated')
        
        try:
            self.sampler.temp = float(new_temp)
        except ValueError:
            pass
        else:
            print('easy_llama: temp updated')

        try:
            self.sampler.frequency_penalty = float(new_frequency_penalty)
        except ValueError:
            pass
        else:
            print('easy_llama: frequency_penalty updated')
        
        try:
            self.sampler.presence_penalty = float(new_presence_penalty)
        except ValueError:
            pass
        else:
            print('easy_llama: presence_penalty updated')
        
        try:
            self.sampler.repeat_penalty = float(new_repeat_penalty)
        except ValueError:
            pass
        else:
            print('easy_llama: repeat_penalty updated')
        
        if _sentinel:   # pretty formatting
            print()
    

    def _interactive_input(
        self,
        prompt: str,
        _dim_style: str,
        _user_style: str,
        _bot_style: str,
        _special_style: str,
        _error_style: str
    ) -> tuple:
        """
        Recive input from the user, while handling multi-line input
        and commands
        """
        full_user_input = '' # may become multiline
        
        while True:
            user_input = input(prompt)
            
            if user_input.endswith('\\'):
                full_user_input += user_input[:-1] + '\n'
            
            elif user_input == '!':

                print()
                try:
                    command = input(f'{RESET_ALL}  ! {_dim_style}')
                except KeyboardInterrupt:
                    print('\n')
                    continue

                if command == '':
                    print(f'\n{_error_style}[no command]{RESET_ALL}\n')

                elif command.lower() in ['reset', 'restart']:
                    self.reset()
                    print(f'\n[thread reset]\n')

                elif command.lower() in ['cls', 'clear']:
                    cls()
                    print()

                elif command.lower() in ['ctx', 'context']:
                    print(f"\n{self.len_messages()}\n")

                elif command.lower() in ['stats', 'print_stats']:
                    print()
                    self.print_stats()
                    print()
                
                elif command.lower() in ['sampler', 'samplers', 'settings']:
                    self._interactive_update_sampler()
                
                elif command.lower() in ['str', 'string', 'as_string']:
                    print(f"\n{self.as_string()}\n")
                
                elif command.lower() in ['repr', 'save', 'backup']:
                    print(f"\n{self!r}\n")
                
                elif command.lower() in ['remove', 'rem', 'delete', 'del']:
                    print()
                    if len(self.messages) == 0:
                        print(f'{_error_style}[no previous message]{RESET_ALL}\n')
                    else:
                        del self.messages[-1]
                        print('[removed last message]\n')

                elif command.lower() in ['last', 'repeat', 'what?']:
                    print()
                    if len(self.messages) == 0:
                        print(f'{_error_style}[no previous message]{RESET_ALL}\n')
                    else:
                        last_msg = self.messages[-1]
                        if last_msg['role'] == 'system':
                            print(f"{_special_style}{last_msg['content']}{RESET_ALL}\n")
                        elif last_msg['role'] == 'user':
                            print(f"{_user_style}{last_msg['content']}{RESET_ALL}\n")
                        elif last_msg['role'] == 'bot':
                            print(f"{_bot_style}{last_msg['content']}{RESET_ALL}\n")
                        else:
                            print(f"{_error_style}{last_msg['content']}{RESET_ALL}\n")
                
                elif command.lower() in ['refill', 'reprint', 'lost', 'find']:
                    cls()
                    print()
                    if len(self.messages) == 0:
                        print(f'{_error_style}[no previous message]{RESET_ALL}\n')
                    else:
                        for msg in self.messages:
                            if msg['role'] == 'system':
                                print(f"{_special_style}{msg['content']}{RESET_ALL}\n")
                            elif msg['role'] == 'user':
                                print(f"{_user_style}{msg['content']}{RESET_ALL}\n")
                            elif msg['role'] == 'bot':
                                print(f"{_bot_style}{msg['content']}{RESET_ALL}\n")
                            else:
                                print(f"{_error_style}{msg['content']}{RESET_ALL}\n")
                        
                elif command.lower() in ['inf', 'inference', 'inf_str']:
                    print(f'\n"""{self.inference_str_from_messages()}"""\n')
                
                elif command.lower() in ['reroll', 're-roll', 're', 'swipe']:
                    old_len = len(self.messages)
                    del self.messages[-1]
                    assert len(self.messages) == (old_len - 1)
                    return '', None
                
                elif command.lower() in ['exit', 'quit']:
                    print(RESET_ALL)
                    return None, None
                
                elif command.lower() in ['sum', 'summ', 'summary', 'summarize']:
                    print('\nGenerating summary...\n')
                    print(self.summarize())
                    print()
                
                elif command.lower() in ['msgs', 'messages', 'msg']:
                    print(f'\n{self.messages}\n')
                
                elif command.lower() in ['help', '/?', '?']:
                    print()
                    print('reset | restart     -- Reset the thread to its original state')
                    print('clear | cls         -- Clear the terminal')
                    print('context | ctx       -- Get the context usage in tokens')
                    print('print_stats | stats -- Get the context usage stats')
                    print('sampler | settings  -- Update the sampler settings')
                    print('string | str        -- Print the message history as a string')
                    print('repr | save         -- Print the representation of the thread')
                    print('remove | delete     -- Remove the last message')
                    print('last | repeat       -- Repeat the last message')
                    print('inference | inf     -- Print the inference string')
                    print('reroll | swipe      -- Regenerate the last message')
                    print('summary | summarize -- Generate a summary of the thread')
                    print('messages | msg      -- Print the raw list of all messages')
                    print('exit | quit         -- Exit the interactive chat (can also use ^C)')
                    print('help | ?            -- Show this screen')
                    print()
                    print("TIP: type < at the prompt and press ENTER to prefix the bot's next message.")
                    print('     for example, type "Sure!" to bypass refusals')
                    print()
                    print("TIP: type !! at the prompt and press ENTER to insert a system message")
                    print()

                else:
                    print(f'\n{_error_style}[unknown command]{RESET_ALL}\n')
            
            # prefix the bot's next message
            elif user_input == '<':

                print()
                try:
                    next_message_start = input(f'{RESET_ALL}  < {_dim_style}')

                except KeyboardInterrupt:
                    print(f'{RESET_ALL}\n')
                    continue

                else:
                    print()
                    return '', next_message_start

            # insert a system message
            elif user_input == '!!':
                print()

                try:
                    next_sys_msg = input(f'{RESET_ALL} !! {_special_style}')
                
                except KeyboardInterrupt:
                    print(f'{RESET_ALL}\n')
                    continue
                
                else:
                    print()
                    return next_sys_msg, -1

            # concatenate multi-line input
            else:
                full_user_input += user_input
                return full_user_input, None


    def interact(
        self,
        color: bool = True,
        header: Optional[str] = None,
        stream: bool = True,
        hook: Optional[Callable] = None,
        prompt: str = '  > '
    ) -> None:
        """
        Start an interactive chat session using this Thread.

        While text is being generated, press `^C` to interrupt the bot.
        Then you have the option to press `ENTER` to re-roll, or to simply type
        another message.

        At the prompt, press `^C` to end the chat session.

        End your input with a backslash `\\` for multi-line input.

        Type `!` and press `ENTER` to enter a basic command prompt. For a list
        of commands, type `help` at this prompt.
        
        Type `<` and press `ENTER` to prefix the bot's next message, for
        example with `Sure!`.

        Type `!!` at the prompt and press `ENTER` to insert a system message.

        The following parameters are optional:
        - color: Whether to use colored text to differentiate user / bot
        - header: Header text to print at the start of the interaction
        - stream: Whether to stream text as it is generated
        """
        print()

        if color:
            from .utils import (
                SPECIAL_STYLE, USER_STYLE,
                BOT_STYLE, DIM_STYLE,
                ERROR_STYLE
            )
        else:
            SPECIAL_STYLE = ''
            USER_STYLE = ''
            BOT_STYLE = ''
            DIM_STYLE = ''
            ERROR_STYLE = ''
        
        if header is not None:
            print(f"{SPECIAL_STYLE}{header}{RESET_ALL}\n")
        
        while True:

            if hook is not None:
                print(DIM_STYLE, end='', flush=True)
                hook(self)
                print(RESET_ALL, end='', flush=True)

            prompt = f"{RESET_ALL}{prompt}{USER_STYLE}"
            
            try:
                user_prompt, next_message_start = self._interactive_input(
                    prompt,
                    DIM_STYLE,
                    USER_STYLE,
                    BOT_STYLE,
                    SPECIAL_STYLE,
                    ERROR_STYLE
                )
            except KeyboardInterrupt:
                print(f"{RESET_ALL}\n")
                return
            
            # got 'exit' or 'quit' command
            if user_prompt is None and next_message_start is None:
                break
            
            # insert a system message via `!!` prompt
            if next_message_start == -1:
                self.add_message('system', user_prompt)
                continue
            
            if next_message_start is not None:
                try:
                    if stream:
                        print(f"{BOT_STYLE}{next_message_start}", end='', flush=True)
                        output = next_message_start + self.model.stream_print(
                            self.inference_str_from_messages() + next_message_start,
                            stops=self.format['stops'],
                            sampler=self.sampler,
                            end=''
                        )
                    else:
                        print(f"{BOT_STYLE}", end='', flush=True)
                        output = next_message_start + self.model.generate(
                            self.inference_str_from_messages() + next_message_start,
                            stops=self.format['stops'],
                            sampler=self.sampler
                        )
                        print(output, end='', flush=True)
                except KeyboardInterrupt:
                    print(f"{DIM_STYLE} [message not added to history; press ENTER to re-roll]\n")
                    continue
                except ExceededContextLengthException as exc:
                    print(f'{ERROR_STYLE}{exc}{RESET_ALL}')
                    print(f'easy_llama: thread: {DIM_STYLE}{self!r}{RESET_ALL}')
                    raise
                else:
                    self.add_message("bot", output)
            else:
                print(BOT_STYLE)
                if user_prompt != "":
                    self.add_message("user", user_prompt)
                try:
                    if stream:
                        output = self.model.stream_print(
                            self.inference_str_from_messages(),
                            stops=self.format['stops'],
                            sampler=self.sampler,
                            end=''
                        )
                    else:
                        output = self.model.generate(
                            self.inference_str_from_messages(),
                            stops=self.format['stops'],
                            sampler=self.sampler
                        )
                        print(output, end='', flush=True)
                except KeyboardInterrupt:
                    print(f"{DIM_STYLE} [message not added to history; press ENTER to re-roll]\n")
                    continue
                except ExceededContextLengthException as exc:
                    print(f'{ERROR_STYLE}{exc}{RESET_ALL}')
                    print(f'easy_llama: thread: {DIM_STYLE}{self!r}{RESET_ALL}')
                    raise
                else:
                    self.add_message("bot", output)

            if output.endswith("\n\n"):
                print(RESET_ALL, end = '', flush=True)
            elif output.endswith("\n"):
                print(RESET_ALL)
            else:
                print(f"{RESET_ALL}\n")


    def reset(self) -> None:
        """
        Reset Thread.messages to its original state
        """
        self.messages: list[Message] = [
            self.create_message("system", self.format['system_prompt'])
        ] if self._messages is None else self._messages
    
    
    def as_string(self) -> str:
        """Return this thread's message history as a string"""
        return ''.join(msg.as_string() for msg in self.messages)

    
    def print_stats(
        self,
        file: _SupportsWriteAndFlush = sys.stdout,
    ) -> None:
        """Print stats about the context usage in this thread"""
        thread_len_tokens = self.len_messages()
        max_ctx_len = self.model.context_length
        c = (thread_len_tokens/max_ctx_len) * 100
        ctx_used_pct = int(c) + (c > int(c)) # round up to next integer
        print(f"{thread_len_tokens} / {max_ctx_len} tokens", file=file)
        print(f"{ctx_used_pct}% of context used", file=file)
        print(f"{len(self.messages)} messages", file=file)
    

    def summarize(
            self,
            messages: Optional[list[Message]] = None,
            model: Optional[Model] = None
        ):
        """
        Generate a summary from a list of messages. If no messages are
        provided, use `self.messages`. If no model is specified, use
        `self.model`.
        """

        assert_type(
            messages,
            (list, NoneType),
            'messages',
            'summarize'
        )

        if model is not None:
            _model = model
        else:
            _model = self.model

        if messages == []:
            raise ValueError(
                f"summarize: the list of messages cannot be empty"
            )

        if messages is None:
            messages = self.messages

        messages_str = ''.join(msg.as_string() for msg in messages)

        inf_str = \
                self.create_message(
                    'system',
                    'Follow the given instructions exactly. Do not add any '
                    'unnecessary information.'
                ).as_string() + \
                self.create_message(
                    'user',
                    'Hello. Take a moment to read the following conversation '
                    'carefully. When you\'re done, write a single paragraph '
                    'that explains all of the most relevant details.'
                    f'\n\n```\n{messages_str}\n```\n\n'
                    'Now that you\'ve read the above conversation, please '
                    'provide a summary in the form of a single paragraph.'
                ).as_string() + \
                self.format['bot_prefix'] + \
                'After carefully reading through the conversation, here\'s' + \
                ' a paragraph that explains the most relevant details:\n\n'

        _model.load()
        summary = _model.generate(
            inf_str,
            stops=self.format['stops'] + ['\n\n'],
            sampler=SamplerSettings(max_len_tokens=512)
        )

        # unload helper model, if used
        if model is not None:
            _model.unload()
        
        return summary


    def warmup(self):
        """
        Ingest the text of this thread into the model's cache
        """
        self.model.ingest(self.inf_str())



# webui.py
# https://github.com/ddh0/easy-llama/

"""Submodule containing the easy-llama WebUI server"""

import os
import sys
import json
import base64

from .utils import (
    InferenceLock,
    print_verbose,
    SPECIAL_STYLE,
    assert_type,
    ERROR_STYLE,
    USER_STYLE,
    BOT_STYLE,
    RESET_ALL
)

from .thread import Thread
from cryptography import x509
from cryptography.x509.oid import NameOID
from datetime import datetime, timedelta, timezone
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from flask import Flask, render_template, request, Response
from cryptography.hazmat.primitives.serialization import Encoding

UTC = timezone.utc

YELLOW = SPECIAL_STYLE
GREEN = USER_STYLE
RED = ERROR_STYLE
RESET = RESET_ALL
BLUE = BOT_STYLE


WARNING = f"""{RED}
################################################################################
{RESET}

                          please keep in mind

        the easy-llama WebUI is not guaranteed to be secure whatsoever

             it is not intended to be exposed to the internet

     if you expose the WebUI to the internet, you do so at your own risk

                          you have been warned

{RED}
################################################################################
{RESET}"""


SSL_CERT_FIRST_TIME_WARNING = \
f"{YELLOW}you have just generated a new self-signed SSL certificate and " + \
f"key. your browser will probably warn you about an untrusted " + \
f"certificate. this is expected and you may safely proceed to the WebUI. " + \
f"subsequent WebUI sessions will re-use this SSL certificate.{RESET}"


ASSETS_FOLDER = os.path.join(os.path.dirname(__file__), 'assets')


MAX_LENGTH_INPUT = 1_000_000 # one million characters


def generate_self_signed_ssl_cert() -> None:
    """
    Generate a self-signed SSL certificate and store it in the assets folder
    """

    private_key = rsa.generate_private_key(
        public_exponent=65537,
        key_size=2048,
    )

    public_key = private_key.public_key()

    # these values are required
    name = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, "XY"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "DUMMY_STATE"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, "DUMMY_LOCALITY"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, "EZLLAMA LLC"),
        x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
    ])

    builder = x509.CertificateBuilder(
        subject_name=name,
        issuer_name=name,
        public_key=public_key,
        serial_number=x509.random_serial_number(),
        not_valid_before=datetime.now(tz=UTC),
        not_valid_after=datetime.now(tz=UTC) + timedelta(days=36500),
    )

    builder = builder.add_extension(
        x509.SubjectAlternativeName([x509.DNSName("localhost")]),
        critical=False,
    )

    # sign the certificate with the private key
    certificate = builder.sign(
        private_key=private_key,
        algorithm=hashes.SHA256(),
    )

    # save key
    with open(f"{ASSETS_FOLDER}/key.pem", "wb") as f:
        f.write(private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        ))

    # save certificate
    with open(f"{ASSETS_FOLDER}/cert.pem", "wb") as f:
        f.write(certificate.public_bytes(Encoding.PEM))


def check_for_ssl_cert() -> bool:
    """
    Return `True` if a local SSL certificate already exists

    If not, remove any orphaned certificate files and return `False`
    """
    fns = [
        f'{ASSETS_FOLDER}/cert.pem',
        f'{ASSETS_FOLDER}/key.pem'
    ]
    if all(os.path.exists(fn) for fn in fns):
        return True
    else:
        for fn in fns:
            if os.path.exists(fn):
                os.remove(fn)
        return False


def newline() -> None:
    """Print a newline to stderr"""
    print('', end='\n', file=sys.stderr, flush=True)


def encode(text: str) -> str:
    """ utf-8 str -> bytes -> ascii base64 """
    data = (text).encode('utf-8')
    encoded = base64.b64encode(data).decode('utf-8')
    #print(f"sending encoded: {encoded!r}")
    return encoded


def decode(text: str) -> str:
    """ ascii base64 -> bytes -> utf-8 str """
    return base64.b64decode(text).decode('utf-8')


def assert_max_length(text: str) -> None:
    """Fail if the given text exceeds 100k characters"""
    if len(text) > MAX_LENGTH_INPUT:
        raise AssertionError(
            'length of input exceeds maximum allowed length of 100k characters'
        )


def _print_inference_string(text: str) -> None:
    print(
        f"{'#' * 80}\n"
        f"{YELLOW}'''{RESET}{text}{YELLOW}'''{RESET}\n"
        f"{'#' * 80}",
        file=sys.stderr,
        flush=True
    )


class WebUI:
    """
    The easy-llama WebUI server
    """

    def __init__(self, thread: Thread):
        """
        Create an instance of the WebUI server based on an existing Thread
        """
        assert_type(thread, Thread, 'thread', 'WebUI')
        self.thread = thread
        self.lock = InferenceLock()
        self._cancel_flag = False

        self.app = Flask(
            __name__,
            static_folder=ASSETS_FOLDER,
            template_folder=ASSETS_FOLDER,
            static_url_path=''
        )

        # variables used for logging to console
        self._log_host = None
        self._log_port = None
    

    def log(self, text: str) -> None:
        if any(i is None for i in [self._log_host, self._log_port]):
            print_verbose(text)
        else:
            print(
                f'easy_llama: WebUI @ '
                f'{YELLOW}{self._log_host}{RESET}:'
                f'{YELLOW}{self._log_port}{RESET}: {text}',
                file=sys.stderr,
                flush=True
            )
        

    def _get_context_string(self) -> str:
        thread_len_tokens = self.thread.len_messages()
        max_ctx_len = self.thread.model.context_length

        return f"{thread_len_tokens} / {max_ctx_len} tokens used"
    

    def start(self, host: str, port: int = 8080, ssl: bool = False):
        """
        Start the WebUI on the specified address and port

        - host `str`:
            The local IP address from which to host the WebUI. For example,
            `'127.0.0.0.1'` or `'10.0.0.140'`
        - port `int`:
            The port on which to host the WebUI. Defaults to `8080`
        - ssl `bool`:
            Whether to generate and use a self-signed SSL certificate to
            encrypt traffic between client and server
        """
        print(WARNING, file=sys.stderr, flush=True)

        assert_type(host, str, 'host', 'WebUI.start')
        assert_type(port, int, 'port', 'WebUI.start')

        self._log_host = None
        self._log_port = None

        self.log(f"starting WebUI instance:")
        self.log(f"   thread.uuid == {self.thread.uuid}")
        self.log(f"   host        == {host}")
        self.log(f"   port        == {port}")
        self.log(f"   ssl (HTTPS) == {ssl}")

        if ssl:

            if check_for_ssl_cert():
                self.log('re-using previously generated SSL certificate')

            else:
                self.log('generating self-signed SSL certifcate')
                generate_self_signed_ssl_cert()
                print_verbose(SSL_CERT_FIRST_TIME_WARNING)


        @self.app.route('/')
        def home():
            return render_template('index.html')
        

        @self.app.route('/convo', methods=['GET'])
        def convo():

            msgs_dict = dict()

            i = 0
            for msg in self.thread.messages:
                msgs_dict[i] = { encode(msg['role']) : encode(msg['content']) }
                i += 1
            
            json_convo = json.dumps(msgs_dict)
            return json_convo, 200, { 'ContentType' : 'application/json' }
        
        
        @self.app.route('/cancel', methods=['POST'])
        def cancel():
            newline()
            self.log('hit cancel endpoint - flag is set')
            self._cancel_flag = True
            return '', 200


        @self.app.route('/submit', methods=['POST'])
        def submit():
            self.log('hit submit endpoint')
            prompt = decode(request.data)

            if prompt in ['', None]:
                self.log('do not submit empty prompt')
                return '', 418

            def generate():
                with self.lock:
                    self.thread.add_message('user', prompt)
                    print(f'{GREEN}{prompt}{RESET}', file=sys.stderr)
                    inf_str = self.thread.inf_str()
                    _print_inference_string(inf_str)
                    token_generator = self.thread.model.stream(
                        inf_str,
                        stops=self.thread.format['stops'],
                        sampler=self.thread.sampler
                    )
                    response = ''

                    for token in token_generator:
                        if self._cancel_flag:
                            print(file=sys.stderr)
                            self.log('cancel generation from /submit. teapot')
                            self._cancel_flag = False
                            return '', 418  # I'm a teapot

                        tok_text = token['choices'][0]['text']
                        response += tok_text
                        print(
                            f'{BLUE}{tok_text}{RESET}',
                            end='',
                            flush=True,
                            file=sys.stderr
                        )
                        yield encode(tok_text) + '\n' # delimiter between tokens

                    self._cancel_flag = False
                    newline()
                    self.thread.add_message('bot', response)

            if prompt not in ['', None]:
                return Response(generate(), mimetype='text/plain')

            return '', 200
        

        @self.app.route('/reset', methods=['POST'])
        def reset():
            self.thread.reset()
            self.log(f"thread was reset")
            return '', 200
        

        @self.app.route('/get_context_string', methods=['GET'])
        def get_context_string():

            json_content = json.dumps(
                { 'text' : encode(self._get_context_string()) }
            )

            return json_content, 200, {'ContentType': 'application/json'}
        

        @self.app.route('/remove', methods=['POST'])
        def remove():

            if len(self.thread.messages) >= 1:
                self.thread.messages.pop(-1)
                self.log('removed last message')
                return '', 200
            
            else:
                self.log('no previous message to remove. teapot')
                return '', 418 # I'm a teapot
        

        @self.app.route('/trigger', methods=['POST'])
        def trigger():
            self.log('hit trigger endpoint')
            prompt = decode(request.data)

            if prompt not in ['', None]:
                self.log(f'trigger with prompt: {prompt!r}')
            else:
                self.log(f'trigger without prompt')
                prompt = ''

            def generate():
                with self.lock:
                    inf_str = self.thread.inf_str() + prompt
                    _print_inference_string(inf_str)
                    token_generator = self.thread.model.stream(
                        inf_str,
                        stops=self.thread.format['stops'],
                        sampler=self.thread.sampler
                    )
                    response = ''

                    for token in token_generator:
                        if self._cancel_flag:
                            print()
                            self.log('cancel generation from /trigger. teapot')
                            self._cancel_flag = False  # reset flag
                            return '', 418  # I'm a teapot

                        tok_text = token['choices'][0]['text']
                        response += tok_text
                        print(f'{BLUE}{tok_text}{RESET}', end='', flush=True)
                        yield encode(tok_text) + '\n' # delimiter between tokens

                    self._cancel_flag = False
                    print('', file=sys.stderr)
                    self.thread.add_message('bot', prompt + response)

            return Response(generate(), mimetype='text/plain')
        

        @self.app.route('/summarize', methods=['GET'])
        def summarize():
            with self.lock:
                summary = self.thread.summarize()
            response = encode(summary)
            self.log(f"generated summary: {BLUE}{summary!r}{RESET}")
            return response, 200, {'ContentType': 'text/plain'}
        

        @self.app.route('/settings', methods=['GET'])
        def settings():
            return render_template('settings.html')


        @self.app.route('/update_sampler', methods=['POST'])
        def update_sampler():
            data = request.get_json()
            self.thread.sampler.max_len_tokens = data.get('max_len_tokens', self.thread.sampler.max_len_tokens)
            self.thread.sampler.top_k = data.get('top_k', self.thread.sampler.top_k)
            self.thread.sampler.top_p = data.get('top_p', self.thread.sampler.top_p)
            self.thread.sampler.min_p = data.get('min_p', self.thread.sampler.min_p)
            self.thread.sampler.temp = data.get('temp', self.thread.sampler.temp)
            self.thread.sampler.frequency_penalty = data.get('frequency_penalty', self.thread.sampler.frequency_penalty)
            self.thread.sampler.presence_penalty = data.get('presence_penalty', self.thread.sampler.presence_penalty)
            self.thread.sampler.repeat_penalty = data.get('repeat_penalty', self.thread.sampler.repeat_penalty)
            return '', 200


        @self.app.route('/get_sampler', methods=['GET'])
        def get_sampler():
            sampler_settings = {
                'max_len_tokens': self.thread.sampler.max_len_tokens,
                'top_k': self.thread.sampler.top_k,
                'top_p': self.thread.sampler.top_p,
                'min_p': self.thread.sampler.min_p,
                'temp': self.thread.sampler.temp,
                'frequency_penalty': self.thread.sampler.frequency_penalty,
                'presence_penalty': self.thread.sampler.presence_penalty,
                'repeat_penalty': self.thread.sampler.repeat_penalty
            }
            return json.dumps(sampler_settings), 200, {'ContentType': 'application/json'}


        if not self.thread.model.is_loaded():

            self.log('loading model')
            self.thread.model.load()

        else:

            self.log('model is already loaded')
        
        self.log('warming up thread')
        self.thread.warmup()

        self.log('now running Flask')

        # these variables are used when printing logs
        self._log_host = host
        self._log_port = port
        
        try:
            self.app.run(
                host=host,
                port=port,
                ssl_context=(
                    f'{ASSETS_FOLDER}/cert.pem',
                    f'{ASSETS_FOLDER}/key.pem'
                ) if ssl else None
            )

        except Exception as exc:
            newline()
            self.log(f'{RED}exception in Flask: {exc}{RESET}')
            raise exc
        
        else:
            newline()
            self._log_host = None
            self._log_port = None
        
        self.log('Flask stopped')



# utils.py
# https://github.com/ddh0/easy-llama/
from ._version import __version__, __llama_cpp_version__

"""Submodule containing convenience functions and QuickGGUFReader"""

import os
import sys
import struct
import numpy as np

from typing    import Iterable, TextIO, Optional
from io        import BufferedReader
from enum      import IntEnum


# color codes used in Thread.interact()
RESET_ALL     = '\x1b[39m'
USER_STYLE    = '\x1b[39m\x1b[32m' # green
BOT_STYLE     = '\x1b[39m\x1b[36m' # blue
DIM_STYLE     = '\x1b[39m\x1b[90m' # grey
SPECIAL_STYLE = '\x1b[39m\x1b[33m' # yellow
ERROR_STYLE   = '\x1b[39m\x1b[91m' # red

NoneType: type = type(None)

class TypeAssertionError(Exception):
    "A call to `assert_type()` has failed"

class _ArrayLike(Iterable):
    "Anything that can be interpreted as a numpy array"

class _SupportsWriteAndFlush(TextIO):
    "A file, stream, or buffer that supports writing and flushing"

class UnreachableException(Exception):
    "The code has reached an unreachable state"
    def __init__(self):
        super().__init__(
            "the code has reached a location that was thought to be "
            "unreachable. please report this issue to the developer at this "
            "link: https://github.com/ddh0/easy-llama/issues/new/choose"
        )

def softmax(
    z: _ArrayLike,
    T: Optional[float] = None,
    dtype: Optional[np.dtype] = None
) -> np.ndarray:
    """
    Compute softmax over values in z, where z is array-like.
    Also apply temperature `T`, if specified.

    Any floating-point value for temperature `T` is valid, including 0.0 and
    negative numbers.

    If `dtype` is not specified, the highest precision numpy `dtype` available
    will be used.
    """
    if dtype is None:
        if hasattr(np, 'float128'):
            _dtype = np.float128
        elif hasattr(np, 'float96'):
            _dtype = np.float96
        elif hasattr(np, 'float80'):
            _dtype = np.float80
        elif hasattr(np, 'float64'):
            _dtype = np.float64
        elif hasattr(np, 'float32'):
            _dtype = np.float32
        elif hasattr(np, 'float16'):
            _dtype = np.float16
        else:
            _dtype = float
    else:
        assert_type(
            dtype,
            type,
            'dtype',
            'softmax',
            'dtype should be a floating type, such as `np.float32`'
        )
        _dtype = dtype
    
    _z = np.asarray(z, dtype=_dtype)
    if T is None or T == 1.0:
        # simple formula with no temperature
        e_z = np.exp(_z - np.max(_z), dtype=_dtype)
        return e_z / np.sum(e_z, axis=0, dtype=_dtype)
    assert_type(T, float, "temperature value 'T'", 'softmax')
    if T == 0.0:
        # Return an array where the maximum value in _z is 1.0 and all others are 0.0
        max_index = np.argmax(_z)
        result = np.zeros_like(_z, dtype=_dtype)
        result[max_index] = 1.0
        return result
    e_z = np.exp(np.divide(_z, T, dtype=_dtype), dtype=_dtype)
    return e_z / np.sum(e_z, axis=0, dtype=_dtype)

def cls() -> None:
    """Clear the terminal"""
    if os.name == 'nt':
        os.system('cls')
    else:
        os.system('clear')
        print("\033c\033[3J", end='', flush=True)

def truncate(text: str) -> str:
    return text if len(text) < 72 else f"{text[:69]}..."

def print_version_info(file: _SupportsWriteAndFlush = sys.stderr) -> None:
    print(f"easy_llama package version: {__version__}", file=file)
    print(f"llama_cpp package version: {__llama_cpp_version__}", file=file)

def print_verbose(text: str) -> None:
    print("easy_llama:", text, file=sys.stderr, flush=True)

def print_info(text: str) -> None:
    print("easy_llama: info:", text, file=sys.stderr, flush=True)

def print_warning(text: str) -> None:
    print(
        f"{RESET_ALL}easy_llama: {SPECIAL_STYLE}WARNING{RESET_ALL}:", text,
        file=sys.stderr, flush=True
    )

def assert_type(
    obj: object,
    expected_type: type | tuple[type],
    obj_name: str,
    code_location: str,
    hint: Optional[str] = None
):
    """
    Ensure that `obj` is an instance of `expected_type`

    If `expected_type` is a tuple, ensure that `obj` is an instance of
    some type in the tuple

    Raise `TypeAssertionError` otherwise, using `obj_name` and
    `code_location` to make an informative exception message

    If specified, `hint` is added as a note to the exception
    """
    if isinstance(obj, expected_type):
        return
    
    obj_type_repr = repr(type(obj).__name__)

    if not isinstance(expected_type, tuple):
        # represent `int` as 'int' instead of "<class 'int'>"
        expected_type_repr = repr(expected_type.__name__)
        exc = TypeAssertionError(
            f"{code_location}: {obj_name} should be an instance of "
            f"{expected_type_repr}, not {obj_type_repr}"
        )
    else:
        # represent `(int, list)` as "('int', 'list')" instead of
        # "(<class 'int'>, <class 'list'>)"
        expected_type_repr = repr(tuple(t.__name__ for t in expected_type))
        exc = TypeAssertionError(
            f"{code_location}: {obj_name} should be one of "
            f"{expected_type_repr}, not {obj_type_repr}"
        )
    if isinstance(hint, str):
        exc.add_note(hint)
    raise exc

def assert_only_ints(iterable: Iterable) -> None:
    """
    Ensure that the given iterable contains only `int`s
    """
    if any(not isinstance(x, int) for x in iterable):
        raise TypeAssertionError(
            f"assert_only_ints: some item in the given iterable is not an int"
        )


class InferenceLock:
    """
    A context manager that can be used to prevent the model from accepting more
    than one generation at a time, which is not supported and can cause a hard
    crash.

    This is mostly useful in asychronous / multi-threaded contexts
    """

    class LockFailure(Exception):
        pass

    def __init__(self):
        self.locked = False

    def __enter__(self):
        return self.acquire()
    
    def __exit__(self, *_):
        return self.release()

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, *_):
       return self.__exit__()
    
    def acquire(self):
        if self.locked:
            raise self.LockFailure(
                'failed to acquire InferenceLock (already locked)'
            )
        self.locked = True
        return self
    
    def release(self):
        if not self.locked:
            raise self.LockFailure(
                'tried to release InferenceLock that is not acquired'
            )
        self.locked = False


class GGUFValueType(IntEnum):
    # Occasionally check to ensure this class is consistent with gguf
    UINT8   = 0
    INT8    = 1
    UINT16  = 2
    INT16   = 3
    UINT32  = 4
    INT32   = 5
    FLOAT32 = 6
    BOOL    = 7
    STRING  = 8
    ARRAY   = 9
    UINT64  = 10
    INT64   = 11
    FLOAT64 = 12

class QuickGGUFReader:
    # ref: https://github.com/ggerganov/ggml/blob/master/docs/gguf.md

    # the GGUF format versions that this class supports
    SUPPORTED_GGUF_VERSIONS = [2, 3]

    # GGUF only supports execution on little or big endian machines
    if sys.byteorder not in ['little', 'big']:
        raise ValueError(
            "host is not little or big endian - GGUF is unsupported"
        )
    
    # arguments for struct.unpack() based on gguf value type
    value_packing: dict = {
        GGUFValueType.UINT8   : "=B",
        GGUFValueType.INT8    : "=b",
        GGUFValueType.UINT16  : "=H",
        GGUFValueType.INT16   : "=h",
        GGUFValueType.UINT32  : "=I",
        GGUFValueType.INT32   : "=i",
        GGUFValueType.FLOAT32 : "=f",
        GGUFValueType.UINT64  : "=Q",
        GGUFValueType.INT64   : "=q",
        GGUFValueType.FLOAT64 : "=d",
        GGUFValueType.BOOL    : "?"
    }

    # length in bytes for each gguf value type
    value_lengths: dict = {
        GGUFValueType.UINT8   : 1,
        GGUFValueType.INT8    : 1,
        GGUFValueType.UINT16  : 2,
        GGUFValueType.INT16   : 2,
        GGUFValueType.UINT32  : 4,
        GGUFValueType.INT32   : 4,
        GGUFValueType.FLOAT32 : 4,
        GGUFValueType.UINT64  : 8,
        GGUFValueType.INT64   : 8,
        GGUFValueType.FLOAT64 : 8,
        GGUFValueType.BOOL    : 1
    }

    @staticmethod
    def unpack(value_type: GGUFValueType, file: BufferedReader):
        return struct.unpack(
            QuickGGUFReader.value_packing.get(value_type),
            file.read(QuickGGUFReader.value_lengths.get(value_type))
        )[0]

    @staticmethod
    def get_single(
            value_type: GGUFValueType,
            file: BufferedReader
        ) -> str | int | float | bool:
        """Read a single value from an open file"""
        if value_type == GGUFValueType.STRING:
            string_length = QuickGGUFReader.unpack(GGUFValueType.UINT64, file=file)
            value = file.read(string_length)
            # officially, strings that cannot be decoded into utf-8 are invalid
            value = value.decode("utf-8")
        else:
            value = QuickGGUFReader.unpack(value_type, file=file)
        return value
    
    @staticmethod
    def load_metadata(
            fn: os.PathLike[str] | str
        ) -> dict[str, str | int | float | bool | list]:
        """
        Given a path to a GGUF file, peek at its header for metadata

        Return a dictionary where all keys are strings, and values can be
        strings, ints, floats, bools, or lists
        """

        metadata: dict[str, str | int | float | bool | list] = {}
        with open(fn, "rb") as file:
            magic = file.read(4)

            if magic != b'GGUF':
                raise ValueError(
                    "your model file is not a valid GGUF file "
                    f"(magic number mismatch, got {magic}, "
                    "expected b'GGUF')"
                )
            
            version = QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)

            if version not in QuickGGUFReader.SUPPORTED_GGUF_VERSIONS:
                raise ValueError(
                    f"your model file reports GGUF version {version}, but "
                    f"only versions {QuickGGUFReader.SUPPORTED_GGUF_VERSIONS} "
                    "are supported. re-convert your model or download a newer "
                    "version"
                )
            
            tensor_count = QuickGGUFReader.unpack(GGUFValueType.UINT64, file=file)
            if version == 3:
                metadata_kv_count = QuickGGUFReader.unpack(GGUFValueType.UINT64, file=file)
            elif version == 2:
                metadata_kv_count = QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)

            for _ in range(metadata_kv_count):
                if version == 3:
                    key_length = QuickGGUFReader.unpack(GGUFValueType.UINT64, file=file)
                elif version == 2:
                    key_length = 0
                    while key_length == 0:
                        # seek until next key is found
                        key_length = QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)
                    file.read(4) # 4 byte offset for GGUFv2
                key = file.read(key_length)
                value_type = GGUFValueType(
                    QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)
                )
                if value_type == GGUFValueType.ARRAY:
                    array_value_type = GGUFValueType(
                        QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)
                    )
                    # array_length is the number of items in the array
                    if version == 3:
                        array_length = QuickGGUFReader.unpack(GGUFValueType.UINT64, file=file)
                    elif version == 2:
                        array_length = QuickGGUFReader.unpack(GGUFValueType.UINT32, file=file)
                        file.read(4) # 4 byte offset for GGUFv2
                    array = [
                        QuickGGUFReader.get_single(
                            array_value_type,
                            file
                        ) for _ in range(array_length)
                    ]
                    metadata[key.decode()] = array
                else:
                    value = QuickGGUFReader.get_single(
                        value_type,
                        file
                    )
                    metadata[key.decode()] = value

        return metadata



# setup.py
# https://github.com/ddh0/easy-llama/
__version__ = '0.1.116'

from setuptools import setup


setup(
    name='easy_llama',
    version=__version__,
    description='Text generation in Python, as easy as possible',
    long_description="For more information, visit the project's GitHub repository:\n\nhttps://github.com/ddh0/easy-llama",
    long_description_content_type='text/markdown',
    url='https://github.com/ddh0/easy-llama/',
    author='Dylan Halladay',
    author_email='dylanhalladay02@icloud.com',
    license='The Unlicense',
    include_package_data=True,
    packages=['easy_llama', 'easy_llama.assets'],
    package_data={
        'easy_llama.assets': [
            '*.ico',
            '*.png',
            '*.html',
            '*.css',
            '*.js',
            '*.webmanifest'
        ]
    },
    install_requires=[
        'llama_cpp_python',
        'numpy',
        'Flask',
        'cryptography'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',  
        'Operating System :: POSIX :: Linux',
        'Natural Language :: English',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ]
)
